# coding: utf-8

import sys
import os
import weakref
import time
import binascii
import os.path as osp
from datetime import datetime
import xml.etree.ElementTree as et

import pdo
from pdo.ipdo_impl import *
from pdo.pdo_impl  import LoginData

# XML API

MSG_CANNOT_GET_LOGIN_DATA            = "Cannot get login data.\n\n{}\n"
MSG_CANNOT_STORE_LOGIN_DATA          = "Cannot store login data.\n\n{}\n"
MSG_CANNOT_REMOVE_LOGIN_DATA         = "Cannot remove login data.\n\n{}\n"
MSG_ERROR_WHILE_CONNECT_TO_DMS       = "Cannot connect to the DMS.\n\n{}\n"
MSG_DATABASE_DOES_NOT_EXISTS         = "Database {} does not exists"
MSG_DATABASE_ALREADY_EXISTS          = "Database {} already exists"
MSG_ERROR_WHILE_LOGIN_TO_DATABASE    = "Cannot login to the database.\n\n{}\n"
MSG_USER_LOGIN_IS_EMPTY              = "User login is empty."
MSG_CANNOT_CHECK_USER_PASSWORD       = "Cannot check password."
MSG_INCORRECT_USER_PASSWORD          = "Incorrect password."
MSG_ERROR_WHILE_LOGOUT_FROM_DATABASE = "Error while logout from the database.\n\n{}\n"
MSG_ERROR_WHILE_BEGIN_TRANSACTION    = "Error while begining transaction.\n\n{}\n"
MSG_ERROR_WHILE_END_TRANSACTION      = "Error while commiting transaction.\n\n{}\n"
MSG_ERROR_WHILE_ABORT_TRANSACTION    = "Error while rollbacking transaction.\n\n{}\n"
MSG_TRANSACTION_NOT_ASSOCIATED       = "Transaction object not associated with database"
MSG_ERROR_WHILE_GETTING_OBJECTS      = "Error while getting objects of {} type.\n\n{}\n"
MSG_ERROR_WHILE_CREATING_OBJECT      = "Error while creating {} '{}'.\n\n{}\n"
MSG_ERROR_WHILE_GETTING_OBJECT       = "Error while getting {} '{}'.\n\n{}\n"
MSG_ERROR_WHILE_GETTING_OBJOFIT      = "Error while getting object of '{}'.\n\n{}\n"
MSG_ERROR_WHILE_DELETING_OBJECT      = "Error while deleting {} '{}'.\n\n{}\n"
MSG_ERROR_WHILE_LOCKING_OBJECT       = "Error while lock operation on {} '{}'.\n\n{}\n"
MSG_ERROR_WHILE_GETTING_PROPERTY     = "Error while getting property '{}' of the {} '{}'.\n\n{}\n"
MSG_ERROR_WHILE_SETTING_PROPERTY     = "Error while setting property '{}' of the {} '{}'.\n\n{}\n"
MSG_ITERATOR_OUT_OF_BOUNDS           = "Iterator out of bounds"
MSG_UNEXP_EXCEPTION                  = "An unexpected error occured: {}"
MSG_NOT_IMPLEMENTED                  = "Not implemented!"

OID_FIELD_NAME  = "id"
ID_CLASS_USER   = "User"
UPWD_FIELD_NAME = "password"


def field2datetime(field):
    try:
        return datetime.strptime(field.strip(), "%Y-%m-%d %H:%M:%S.%f")
    except ValueError:
        return datetime.strptime(field.strip(), "%Y-%m-%d %H:%M:%S")


def datetime2field(pdate):
    if pdate.tzinfo is not None:
        # Convert pdate to UTC
        pdate = (pdate - pdate.utcoffset()).replace(tzinfo=None)
    return pdate.isoformat(" ")


def field2bitmask(field):
    return int(field)


def bitmask2field(bitmask):
    return int(bitmask)


#-----------------------------------------------------------------------------#
# XML access routines                                                         #
#-----------------------------------------------------------------------------#


def NewDatabaseService(db_source="", options={}):

    DatabaseService.init()

    service = pdo.DatabaseService()
    service._impl = DatabaseService(db_source, dict(options))
    return service


@interface.implementer(IDatabaseService)
class DatabaseService:

    ID = 4

    @staticmethod
    def initConnection():
        pass

    @staticmethod
    def init():
        pass

    def __init__(self, db_source, options):

        self.source  = db_source
        self.options = options

    def db_source(self):

        return self.source

    def get_login_data(self):

        DatabaseService.initConnection()

        login_data = LoginData()

        try:
            pass # raise NotImplementedError(MSG_NOT_IMPLEMENTED)
        except Exception as exc:
            error = MSG_UNEXP_EXCEPTION.format(exc)
            raise pdo.XConnectError(MSG_CANNOT_GET_LOGIN_DATA.format(error))
        return login_data

    def store_login_data(self, login_data):

        DatabaseService.initConnection()

        try:
            pass # raise NotImplementedError(MSG_NOT_IMPLEMENTED)
        except Exception as exc:
            error = MSG_UNEXP_EXCEPTION.format(exc)
            raise pdo.XConnectError(MSG_CANNOT_STORE_LOGIN_DATA.format(error))

    def remove_login_data(self):

        DatabaseService.initConnection()

        try:
            pass # raise NotImplementedError(MSG_NOT_IMPLEMENTED)
        except Exception as exc:
            error = MSG_UNEXP_EXCEPTION.format(exc)
            raise pdo.XConnectError(MSG_CANNOT_REMOVE_LOGIN_DATA.format(error))

    def get_db_names(self):

        DatabaseService.initConnection()

        names = {}
        try:
            xml_DOM = et.parse(self.source)
            for node in xml_DOM.findall("Database[@db_name]"):
                name = node.get("db_name")
                if name:
                    options = {}
                    names[str(name)] = options
        except Exception as exc:
            raise pdo.XConnectError(MSG_ERROR_WHILE_CONNECT_TO_DMS.format(exc))
        return names

    def new_db(self, db_name, db_options):

        Database.init()
        DatabaseService.initConnection()

        database = Database(db_name, db_options)
        database.service_ref = weakref.ref(self)

        try:
            xml_DOM = et.parse(self.source)
            db_node = xml_DOM.find("Database[@db_name='{}']".format(db_name))
        except IOError as exc:
            pass
        except Exception as exc:
            raise pdo.XConnectError(MSG_ERROR_WHILE_CONNECT_TO_DMS.format(exc))
        if db_node is not None:
            error = MSG_DATABASE_ALREADY_EXISTS.format(db_name)
            raise pdo.XConnectError(MSG_ERROR_WHILE_CONNECT_TO_DMS.format(error))

        try:
            root    = et.Element("DMS")
            db_node = et.SubElement(root, "Database")
            db_node.set("db_name", db_name)
            xml_DOM = et.ElementTree(root)
        except Exception as exc:
            error = MSG_UNEXP_EXCEPTION.format(exc)
            raise pdo.XConnectError(MSG_ERROR_WHILE_CONNECT_TO_DMS.format(error))

        database.xml_DOM = xml_DOM
        database.db_node = db_node
        return database

    def open_db(self, db_name, db_options):

        Database.init()
        DatabaseService.initConnection()

        database = Database(db_name, db_options)
        database.service_ref = weakref.ref(self)

        try:
            xml_DOM = et.parse(self.source)
            db_node = xml_DOM.find("Database[@db_name='{}']".format(db_name))
        except Exception as exc:
            raise pdo.XConnectError(MSG_ERROR_WHILE_CONNECT_TO_DMS.format(exc))
        if db_node is None:
            error = MSG_DATABASE_DOES_NOT_EXISTS.format(db_name)
            raise pdo.XConnectError(MSG_ERROR_WHILE_CONNECT_TO_DMS.format(error))

        database.xml_DOM = xml_DOM
        database.db_node = db_node
        return database


@interface.implementer(IDatabase)
class Database:

    @staticmethod
    def init():
        pass

    def __init__(self, db_name, db_options):

        self.name     = db_name
        self.options  = db_options
        self.user     = ""
        self.password = ""

        self.service_ref = None
        self.connected   = False

        self.xml_DOM = None
        self.db_node = None

    def __del__(self):

        try:
            self.logout()
        except:
            pass
        try:
            if self.xml_DOM is not None:
                self.xml_DOM.write(self.service_ref().source)
                #!!! roboczo tu, ale dorobic cos w rodzaju funkcji sync !!!
        except Exception as exc:
            raise pdo.XConnectError(MSG_ERROR_WHILE_CONNECT_TO_DMS.format(exc))

    def db_name(self):

        return self.name

    def db_options(self):

        return self.options

    def user_name(self):

        return self.user

    def user_password(self):

        return self.password

    def login(self, user_name, user_password):

        if not user_name:
            error = MSG_USER_LOGIN_IS_EMPTY
            raise pdo.XOpenError(MSG_ERROR_WHILE_LOGIN_TO_DATABASE.format(error))

        try:
            user_obj = self.get_object(ID_CLASS_USER, user_name)
        except Exception as exc:
            raise pdo.XOpenError(MSG_ERROR_WHILE_LOGIN_TO_DATABASE.format(exc))

        try:
            db_password = user_obj.getBytes(UPWD_FIELD_NAME)
        except Exception as exc:
            error = MSG_CANNOT_CHECK_USER_PASSWORD
            raise pdo.XOpenError(MSG_ERROR_WHILE_LOGIN_TO_DATABASE.format(error))

        del user_obj

        from hashlib import sha224
        user_password_crypted = sha224(user_password.encode("utf-8")).hexdigest().encode("utf-8")

        if db_password and user_password_crypted != db_password:
            error = MSG_INCORRECT_USER_PASSWORD
            raise pdo.XOpenError(MSG_ERROR_WHILE_LOGIN_TO_DATABASE.format(error))

        self.user     = user_name
        self.password = user_password

    def logout(self):

        self.user     = ""
        self.password = ""

    def get_transaction(self):

        tx = Transaction()
        tx.db = self  # !!! weakref ? !!!
        return tx

    def get_object(self, obj_type, obj_id):

        try:
            type_tag   = str(obj_type).replace(" ", "_")
            folder_tag = type_tag + "_folder"
            xml_elem = self.db_node.find(folder_tag+"/"+type_tag+"[@{}='{}']".format(OID_FIELD_NAME, obj_id))
        except Exception as exc:
            error = MSG_UNEXP_EXCEPTION.format(exc)
            raise pdo.XError(MSG_ERROR_WHILE_GETTING_OBJECT.format(
                             obj_type, obj_id, error))
        if xml_elem is None:
            error = "???"
            raise pdo.XDataError(MSG_ERROR_WHILE_GETTING_OBJECT.format(
                                 obj_type, obj_id, error))

        obj = Object(obj_type, obj_id)
        obj.db_node  = self.db_node
        obj.xml_elem = xml_elem
        return obj

    def new_object(self, obj_type, obj_id):

        try:
            type_tag   = str(obj_type).replace(" ", "_")
            folder_tag = type_tag + "_folder"
            xml_elem = self.db_node.find(folder_tag+"/"+type_tag+"[@{}='{}']".format(OID_FIELD_NAME, obj_id))
        except Exception as exc:
            error = MSG_UNEXP_EXCEPTION.format(exc)
            raise pdo.XError(MSG_ERROR_WHILE_CREATING_OBJECT.format(
                             obj_type, obj_id, error))
        if xml_elem is not None:
            error = "???"
            raise pdo.XDataError(MSG_ERROR_WHILE_GETTING_OBJECT.format(
                                 obj_type, obj_id, error))
        try:

            parent = self.db_node.find(folder_tag)
            if parent is None:
                parent = et.SubElement(self.db_node, folder_tag)
            xml_elem = et.SubElement(parent, type_tag)
            xml_elem.set(OID_FIELD_NAME, str(obj_id))

        except Exception as exc:
            error = MSG_UNEXP_EXCEPTION.format(exc)
            raise pdo.XError(MSG_ERROR_WHILE_CREATING_OBJECT.format(
                             obj_type, obj_id, error))

        obj = Object(obj_type, obj_id)
        obj.db_node  = self.db_node
        obj.xml_elem = xml_elem
        return obj

    def get_object_iterator(self, obj_type, restrictions):

        it = ObjectIterator(obj_type, restrictions)
        it.db = self  # !!! weakref ? !!!

        it.first()
        return it


@interface.implementer(ITransaction)
class Transaction:

    def __init__(self):

        self.db = None

    def begin(self):

        if self.db is None:
            error = MSG_TRANSACTION_NOT_ASSOCIATED
            raise pdo.XError(MSG_ERROR_WHILE_BEGIN_TRANSACTION.format(error))

        try:
            pass # raise NotImplementedError(MSG_NOT_IMPLEMENTED)
        except Exception as exc:
            error = MSG_UNEXP_EXCEPTION.format(exc)
            raise pdo.XError(MSG_ERROR_WHILE_BEGIN_TRANSACTION.format(error))

    def commit(self):

        if self.db is None:
            error = MSG_TRANSACTION_NOT_ASSOCIATED
            raise pdo.XError(MSG_ERROR_WHILE_END_TRANSACTION.format(error))

        try:
            pass # raise NotImplementedError(MSG_NOT_IMPLEMENTED)
        except Exception as exc:
            error = MSG_UNEXP_EXCEPTION.format(exc)
            raise pdo.XError(MSG_ERROR_WHILE_END_TRANSACTION.format(error))

    def rollback(self):

        if self.db is None:
            try:
                pass # raise NotImplementedError(MSG_NOT_IMPLEMENTED)
            except:
                pass
            error = MSG_TRANSACTION_NOT_ASSOCIATED
            raise pdo.XError(MSG_ERROR_WHILE_ABORT_TRANSACTION.format(error))

        try:
            pass # raise NotImplementedError(MSG_NOT_IMPLEMENTED)
        except Exception as exc:
            error = MSG_UNEXP_EXCEPTION.format(exc)
            raise pdo.XError(MSG_ERROR_WHILE_ABORT_TRANSACTION.format(error))


@interface.implementer(IObject)
class Object:

    def __init__(self, obj_type, obj_id):

        self.obj_type = obj_type
        self.obj_id   = obj_id

        self.db_node  = None
        self.xml_elem = None

    def get_type(self):

        return self.obj_type

    def getString(self, prop_id):

        try:
            value = self.xml_elem.get(prop_id)
        except Exception as exc:
            self._get_handle_exc(prop_id, exc)
        else:
            self._validate_value(prop_id, value)
        return str(value)

    def getBytes(self, prop_id):

        try:
            value = self.xml_elem.get(prop_id)
        except Exception as exc:
            self._get_handle_exc(prop_id, exc)
        else:
            self._validate_value(prop_id, value)
        return str(value).encode("ASCII")

    def getBoolean(self, prop_id):

        try:
            value = self.xml_elem.get(prop_id)
        except Exception as exc:
            self._get_handle_exc(prop_id, exc)
        else:
            self._validate_value(prop_id, value)
        return bool(value)

    def getInteger(self, prop_id):

        try:
            value = self.xml_elem.get(prop_id)
        except Exception as exc:
            self._get_handle_exc(prop_id, exc)
        else:
            self._validate_value(prop_id, value)
        return int(value)

    def getFloat(self, prop_id):

        try:
            value = self.xml_elem.get(prop_id)
        except Exception as exc:
            self._get_handle_exc(prop_id, exc)
        else:
            self._validate_value(prop_id, value)
        return float(value)

    def getTimestamp(self, prop_id):

        try:
            value = self.xml_elem.get(prop_id)
        except Exception as exc:
            self._get_handle_exc(prop_id, exc)
        else:
            self._validate_value(prop_id, value)
        return field2datetime(value)

    def getBitmask(self, prop_id):

        try:
            value = self.xml_elem.get(prop_id)
        except Exception as exc:
            self._get_handle_exc(prop_id, exc)
        else:
            self._validate_value(prop_id, value)
        return field2bitmask(value)

    def getBlob(self, prop_id):

        try:
            value = self.xml_elem.get(prop_id)
        except Exception as exc:
            self._get_handle_exc(prop_id, exc)
        else:
            self._validate_value(prop_id, value)
        return binascii.a2b_hex(str(value).encode("ASCII"))

    def checkoutBlob(self, prop_id, path):

        try:
            value = self.xml_elem.get(prop_id)
        except Exception as exc:
            self._get_handle_exc(prop_id, exc)
        else:
            self._validate_value(prop_id, value)

        try:
            buflen = 128 * 1024
            with open(osp.normpath(path), "wb", buflen) as fo:
                fo.write(binascii.a2b_hex(str(value).encode("ASCII")))
        except Exception as exc:
            raise pdo.XError(MSG_ERROR_WHILE_GETTING_PROPERTY.format(
                             prop_id, self.obj_type, self.obj_id, exc))

    def getListIterator(self, prop_id):

        it = ListIterator(prop_id)
        it.obj_type = self.obj_type
        it.obj_id   = self.obj_id
        it.xml_elem = self.xml_elem  # !!! weakref ? !!!

        it.first()
        return it

    def setString(self, prop_id, value):

        # !!! dopisac transakcyjnosc, gdy nie ma ogolnej transakcji !!!

        try:
            self.xml_elem.set(prop_id, str(value))
        except Exception as exc:
            self._set_handle_exc(prop_id, exc)

    def setBytes(self, prop_id, value):

        # !!! dopisac transakcyjnosc, gdy nie ma ogolnej transakcji !!!

        try:
            self.xml_elem.set(prop_id, str(value).encode("ASCII"))
        except Exception as exc:
            self._set_handle_exc(prop_id, exc)

    def setBoolean(self, prop_id, value):

        # !!! dopisac transakcyjnosc, gdy nie ma ogolnej transakcji !!!

        try:
            self.xml_elem.set(prop_id, int(bool(value)))
        except Exception as exc:
            self._set_handle_exc(prop_id, exc)

    def setInteger(self, prop_id, value):

        # !!! dopisac transakcyjnosc, gdy nie ma ogolnej transakcji !!!

        try:
            self.xml_elem.set(prop_id, int(value))
        except Exception as exc:
            self._set_handle_exc(prop_id, exc)

    def setFloat(self, prop_id, value):

        # !!! dopisac transakcyjnosc, gdy nie ma ogolnej transakcji !!!

        try:
            self.xml_elem.set(prop_id, float(value))
        except Exception as exc:
            self._set_handle_exc(prop_id, exc)

    def setTimestamp(self, prop_id, value):

        # !!! dopisac transakcyjnosc, gdy nie ma ogolnej transakcji !!!

        try:
            self.xml_elem.set(prop_id, datetime2field(value))
        except Exception as exc:
            self._set_handle_exc(prop_id, exc)

    def setBitmask(self, prop_id, value): # unsigned long

        # !!! dopisac transakcyjnosc, gdy nie ma ogolnej transakcji !!!

        try:
            self.xml_elem.set(prop_id, bitmask2field(value))
        except Exception as exc:
            self._set_handle_exc(prop_id, exc)

    def setBlob(self, prop_id, value): # string

        # !!! dopisac transakcyjnosc, gdy nie ma ogolnej transakcji !!!

        try:
            self.xml_elem.set(prop_id, binascii.b2a_hex(value))
        except Exception as exc:
            self._set_handle_exc(prop_id, exc)

    def checkinBlob(self, prop_id, path): # path

        # !!! dopisac transakcyjnosc, gdy nie ma ogolnej transakcji !!!

        try:
            buflen = 128 * 1024
            with open(path, "rb", buflen) as fi:
                value = fi.read()
        except Exception as exc:
            raise pdo.XError(MSG_ERROR_WHILE_SETTING_PROPERTY.format(
                             prop_id, self.obj_type, self.obj_id, exc))
        try:
            self.xml_elem.set(prop_id, binascii.b2a_hex(value))
        except Exception as exc:
            self._set_handle_exc(prop_id, exc)

    def update(self):

        pass

    def remove(self):

        try:
            type_tag   = str(self.obj_type).replace(" ", "_")
            folder_tag = type_tag + "_folder"
            for parent in self.db_node.findall(folder_tag):
                if et.iselement(parent):
                    try:
                        parent.remove(self.xml_elem)
                    except ValueError as exc:
                        pass
            self.xml_elem.clear()
            self.xml_elem = None
        except Exception as exc:
            error = MSG_UNEXP_EXCEPTION.format(exc)
            raise pdo.XError(MSG_ERROR_WHILE_DELETING_OBJECT.format(
                             self.obj_type, self.obj_id, error))

    def lock(self):

        try:
            raise NotImplementedError(MSG_NOT_IMPLEMENTED)
        except Exception as exc:
            self._lock_handle_exc(exc)

    def unlock(self):

        try:
            raise NotImplementedError(MSG_NOT_IMPLEMENTED)
        except Exception as exc:
            self._lock_handle_exc(exc)

    def locked(self):

        try:
            return False  # !!! ew dopisac !!!
        except Exception as exc:
            self._lock_handle_exc(exc)

    def own_lock(self):

        try:
            return False  # !!! ew dopisac !!!
        except Exception as exc:
            self._lock_handle_exc(exc)

    def outer_lock(self):

        return self.locked() and not self.own_lock()

    def locker(self):

        try:
            raise NotImplementedError(MSG_NOT_IMPLEMENTED)
            #return locked_by
        except Exception as exc:
            self._lock_handle_exc(exc)

    def _get_handle_exc(self, prop_id, exc):

        error = MSG_UNEXP_EXCEPTION.format(exc)
        raise pdo.XError(MSG_ERROR_WHILE_GETTING_PROPERTY.format(
                         prop_id, self.obj_type, self.obj_id, error))

    def _validate_value(self, prop_id, value):

        if value is None:
            error = "???"
            raise pdo.XDataError(MSG_ERROR_WHILE_GETTING_PROPERTY.format(
                                 prop_id, self.obj_type, self.obj_id, error))

    def _set_handle_exc(self, prop_id, exc):

        error = MSG_UNEXP_EXCEPTION.format(exc)
        raise pdo.XError(MSG_ERROR_WHILE_SETTING_PROPERTY.format(
                         prop_id, self.obj_type, self.obj_id, error))

    def _lock_handle_exc(self, exc):

        error = MSG_UNEXP_EXCEPTION.format(exc)
        raise pdo.XError(MSG_ERROR_WHILE_LOCKING_OBJECT.format(
                         self.obj_type, self.obj_id, error))


@interface.implementer(IObjectIterator)
class ObjectIterator:

    def __init__(self, obj_type, restrictions):

        self.obj_type     = obj_type
        self.restrictions = restrictions

        self.db = None  # !!! weakref ? !!!

        self.item_iter = None
        self.has_item  = False
        self.item_row  = None
        self._follow   = True

    def __iter__(self):

        return self

    def __bool__(self):

        return self.has_item

    def first(self):

        type_tag   = str(self.obj_type).replace(" ", "_")
        folder_tag = type_tag + "_folder"
        self.item_iter = iter(self.db.db_node.findall(folder_tag+"/"+type_tag))

        self._follow = True
        self._step()
        self._follow = False

    def next(self):

        self._step()
        self._follow = True

        if not self.has_item:
            raise StopIteration

        return self

    def _step(self):

        if self._follow:
            self.item_row = None
            for item in self.item_iter:
                if et.iselement(item):
                    self.item_row = item
                    break
            self.has_item = self.item_row is not None

    def _oid(self):

        oid = "(unknown)"
        try:
            obj_id = self.item_row.get(OID_FIELD_NAME)
            if obj_id is not None:
                oid = str(obj_id)
        except:
            pass
        return oid

    def getString(self, prop_id):

        self._get_precheck()
        try:
            value = self.item_row.get(prop_id)
        except Exception as exc:
            self._get_handle_exc(prop_id, exc)
        else:
            self._validate_value(prop_id, value)
        return str(value)

    def getBytes(self, prop_id):

        self._get_precheck()
        try:
            value = self.item_row.get(prop_id)
        except Exception as exc:
            self._get_handle_exc(prop_id, exc)
        else:
            self._validate_value(prop_id, value)
        return str(value).encode("ASCII")

    def getBoolean(self, prop_id):

        self._get_precheck()
        try:
            value = self.item_row.get(prop_id)
        except Exception as exc:
            self._get_handle_exc(prop_id, exc)
        else:
            self._validate_value(prop_id, value)
        return bool(value)

    def getInteger(self, prop_id):

        self._get_precheck()
        try:
            value = self.item_row.get(prop_id)
        except Exception as exc:
            self._get_handle_exc(prop_id, exc)
        else:
            self._validate_value(prop_id, value)
        return int(value)

    def getFloat(self, prop_id):

        self._get_precheck()
        try:
            value = self.item_row.get(prop_id)
        except Exception as exc:
            self._get_handle_exc(prop_id, exc)
        else:
            self._validate_value(prop_id, value)
        return float(value)

    def getTimestamp(self, prop_id):

        self._get_precheck()
        try:
            value = self.item_row.get(prop_id)
        except Exception as exc:
            self._get_handle_exc(prop_id, exc)
        else:
            self._validate_value(prop_id, value)
        return field2datetime(value)

    def getBitmask(self, prop_id):

        self._get_precheck()
        try:
            value = self.item_row.get(prop_id)
        except Exception as exc:
            self._get_handle_exc(prop_id, exc)
        else:
            self._validate_value(prop_id, value)
        return field2bitmask(value)

    def getBlob(self, prop_id):

        self._get_precheck()
        try:
            value = self.item_row.get(prop_id)
        except Exception as exc:
            self._get_handle_exc(prop_id, exc)
        else:
            self._validate_value(prop_id, value)
        return binascii.a2b_hex(str(value).encode("ASCII"))

    def checkoutBlob(self, prop_id, path):

        self._get_precheck()
        try:
            value = self.item_row.get(prop_id)
        except Exception as exc:
            self._get_handle_exc(prop_id, exc)
        else:
            self._validate_value(prop_id, value)
        try:
            buflen = 128 * 1024
            with open(osp.normpath(path), "wb", buflen) as fo:
                fo.write(binascii.a2b_hex(str(value).encode("ASCII")))
        except Exception as exc:
            raise pdo.XError(MSG_ERROR_WHILE_GETTING_PROPERTY.format(
                             prop_id, self.obj_type, self._oid(), exc))

    def get_object(self):

        self._get_precheck()
        try:
            obj_id = self.item_row.get(OID_FIELD_NAME)
        except Exception as exc:
            error = MSG_UNEXP_EXCEPTION.format(exc)
            raise pdo.XError(MSG_ERROR_WHILE_GETTING_OBJOFIT.format(
                             self.obj_type, error))
        if obj_id is None:
            error = "???"
            raise pdo.XError(MSG_ERROR_WHILE_GETTING_OBJOFIT.format(
                             self.obj_type, error))
        obj = Object(self.obj_type, obj_id)
        obj.db_node  = self.db.db_node
        obj.xml_elem = self.item_row
        return obj

    def _get_precheck(self):

        if not self.has_item:
            error = MSG_ITERATOR_OUT_OF_BOUNDS
            raise pdo.XError(MSG_ERROR_WHILE_GETTING_OBJOFIT.format(
                             self.obj_type, error))

    def _get_handle_exc(self, prop_id, exc):

        error = MSG_UNEXP_EXCEPTION.format(exc)
        raise pdo.XError(MSG_ERROR_WHILE_GETTING_PROPERTY.format(
                         prop_id, self.obj_type, self._oid(), error))

    def _validate_value(self, prop_id, value):

        if value is None:
            error = "???"
            raise pdo.XDataError(MSG_ERROR_WHILE_GETTING_PROPERTY.format(
                                 prop_id, self.obj_type, self._oid(), error))


@interface.implementer(IListIterator)
class ListIterator:

    def __init__(self, prop_id):

        self.obj_type = None
        self.obj_id   = None
        self.prop_id  = prop_id
        self.xml_elem = None  # !!! weakref ? !!!

        self.item_iter = None
        self.has_item  = False
        self.item_row  = None
        self._follow   = True

    def __iter__(self):

        return self

    def __bool__(self):

        return self.has_item

    def first(self):

        parent = self.xml_elem.find(self.prop_id)
        if parent is None:
            parent = []
        self.item_iter = iter(parent)

        self._follow = True
        self._step()
        self._follow = False

    def next(self):

        self._step()
        self._follow = True

        if not self.has_item:
            raise StopIteration

        return self

    def _step(self):

        if self._follow:
            self.item_row = next((item for item in self.item_iter
                                  if et.iselement(item)), None)
            self.has_item = self.item_row is not None

    def add(self):

        try:
            parent = self.xml_elem.find(self.prop_id)
            if parent is None:
                parent = et.SubElement(self.xml_elem, self.prop_id)
            self.item_row = et.SubElement(parent, "item")
        except Exception as exc:
            error = MSG_UNEXP_EXCEPTION.format(exc)
            raise pdo.XError(MSG_ERROR_WHILE_SETTING_PROPERTY.format(
                             self.prop_id, self.obj_type, self.obj_id, error))

    def remove(self):

        if not self.has_item:
            return

        try:
            parent = self.xml_elem.find(self.prop_id)

            parent.remove(self.item_row)
            self.item_row.clear()

            try:
                self._follow = True
                self.next()
            except StopIteration:
                pass
        except Exception as exc:
            error = MSG_UNEXP_EXCEPTION.format(exc)
            raise pdo.XError(MSG_ERROR_WHILE_SETTING_PROPERTY.format(
                             self.prop_id, self.obj_type, self.obj_id, error))

    def clear(self):

        self.first()
        while self.has_item:
            self.remove()

    def getString(self, prop_id):

        self._get_precheck(prop_id)
        try:
            value = self.item_row.get(prop_id)
        except Exception as exc:
            self._get_handle_exc(prop_id, exc)
        else:
            self._validate_value(prop_id, value)
        return str(value)

    def getBytes(self, prop_id):

        self._get_precheck(prop_id)
        try:
            value = self.item_row.get(prop_id)
        except Exception as exc:
            self._get_handle_exc(prop_id, exc)
        else:
            self._validate_value(prop_id, value)
        return str(value).encode("ASCII")

    def getBoolean(self, prop_id):

        self._get_precheck(prop_id)
        try:
            value = self.item_row.get(prop_id)
        except Exception as exc:
            self._get_handle_exc(prop_id, exc)
        else:
            self._validate_value(prop_id, value)
        return bool(value)

    def getInteger(self, prop_id):

        self._get_precheck(prop_id)
        try:
            value = self.item_row.get(prop_id)
        except Exception as exc:
            self._get_handle_exc(prop_id, exc)
        else:
            self._validate_value(prop_id, value)
        return int(value)

    def getFloat(self, prop_id):

        self._get_precheck(prop_id)
        try:
            value = self.item_row.get(prop_id)
        except Exception as exc:
            self._get_handle_exc(prop_id, exc)
        else:
            self._validate_value(prop_id, value)
        return float(value)

    def getTimestamp(self, prop_id):

        self._get_precheck(prop_id)
        try:
            value = self.item_row.get(prop_id)
        except Exception as exc:
            self._get_handle_exc(prop_id, exc)
        else:
            self._validate_value(prop_id, value)
        return field2datetime(value)

    def getBitmask(self, prop_id):

        self._get_precheck(prop_id)
        try:
            value = self.item_row.get(prop_id)
        except Exception as exc:
            self._get_handle_exc(prop_id, exc)
        else:
            self._validate_value(prop_id, value)
        return field2bitmask(value)

    def getBlob(self, prop_id):

        self._get_precheck(prop_id)
        try:
            value = self.item_row.get(prop_id)
        except Exception as exc:
            self._get_handle_exc(prop_id, exc)
        else:
            self._validate_value(prop_id, value)
        return binascii.a2b_hex(str(value).encode("ASCII"))

    def checkoutBlob(self, prop_id, path):

        self._get_precheck(prop_id)
        try:
            value = self.item_row.get(prop_id)
        except Exception as exc:
            self._get_handle_exc(prop_id, exc)
        else:
            self._validate_value(prop_id, value)
        try:
            buflen = 128 * 1024
            with open(osp.normpath(path), "wb", buflen) as fo:
                fo.write(binascii.a2b_hex(str(value).encode("ASCII")))
        except Exception as exc:
            raise pdo.XError(MSG_ERROR_WHILE_GETTING_PROPERTY.format(
                             prop_id, self.obj_type, self.obj_id, exc))

    def getListIterator(self, prop_id):

        self._get_precheck(prop_id)

        it = ListIterator(self.prop_id + "/" + prop_id)
        it.obj_type = self.obj_type
        it.obj_id   = self.obj_id
        it.xml_elem = self.xml_elem  # !!! weakref ? !!!

        it.first()
        return it

    def setString(self, prop_id, value):

        self._set_precheck(prop_id)
        try:
            self.item_row.set(prop_id, str(value))
        except Exception as exc:
            self._set_handle_exc(prop_id, exc)

    def setBytes(self, prop_id, value):

        self._set_precheck(prop_id)
        try:
            self.item_row.set(prop_id, str(value).encode("ASCII"))
        except Exception as exc:
            self._set_handle_exc(prop_id, exc)

    def setBoolean(self, prop_id, value):

        self._set_precheck(prop_id)
        try:
            self.item_row.set(prop_id, int(bool(value)))
        except Exception as exc:
            self._set_handle_exc(prop_id, exc)

    def setInteger(self, prop_id, value):

        self._set_precheck(prop_id)
        try:
            self.item_row.set(prop_id, int(value))
        except Exception as exc:
            self._set_handle_exc(prop_id, exc)

    def setFloat(self, prop_id, value):

        self._set_precheck(prop_id)
        try:
            self.item_row.set(prop_id, float(value))
        except Exception as exc:
            self._set_handle_exc(prop_id, exc)

    def setTimestamp(self, prop_id, value):

        self._set_precheck(prop_id)
        try:
            self.item_row.set(prop_id, datetime2field(value))
        except Exception as exc:
            self._set_handle_exc(prop_id, exc)

    def setBitmask(self, prop_id, value): # unsigned long

        self._set_precheck(prop_id)
        try:
            self.item_row.set(prop_id, bitmask2field(value))
        except Exception as exc:
            self._set_handle_exc(prop_id, exc)

    def setBlob(self, prop_id, value):

        self._set_precheck(prop_id)
        try:
            self.item_row.set(prop_id, binascii.b2a_hex(value))
        except Exception as exc:
            self._set_handle_exc(prop_id, exc)

    def checkinBlob(self, prop_id, path):

        self._set_precheck(prop_id)
        try:
            buflen = 128 * 1024
            with open(path, "rb", buflen) as fi:
                value = fi.read()
        except Exception as exc:
            raise pdo.XError(MSG_ERROR_WHILE_SETTING_PROPERTY.format(
                             prop_id, self.obj_type, self.obj_id, exc))
        try:
            self.item_row.set(prop_id, binascii.b2a_hex(value))
        except Exception as exc:
            self._set_handle_exc(prop_id, exc)

    def _get_precheck(self, prop_id):

        if not self.has_item:
            error = MSG_ITERATOR_OUT_OF_BOUNDS
            raise pdo.XError(MSG_ERROR_WHILE_GETTING_PROPERTY.format(
                             prop_id, self.obj_type, self.obj_id, error))

    def _get_handle_exc(self, prop_id, exc):

        error = MSG_UNEXP_EXCEPTION.format(exc)
        raise pdo.XError(MSG_ERROR_WHILE_GETTING_PROPERTY.format(
                         prop_id, self.obj_type, self.obj_id, error))

    def _validate_value(self, prop_id, value):

        if value is None:
            error = "???"
            raise pdo.XDataError(MSG_ERROR_WHILE_GETTING_PROPERTY.format(
                                 prop_id, self.obj_type, self.obj_id, error))

    def _set_precheck(self, prop_id):

        if not self.has_item:
            error = MSG_ITERATOR_OUT_OF_BOUNDS
            raise pdo.XError(MSG_ERROR_WHILE_SETTING_PROPERTY.format(
                             prop_id, self.obj_type, self.obj_id, error))

    def _set_handle_exc(self, prop_id, exc):

        error = MSG_UNEXP_EXCEPTION.format(exc)
        raise pdo.XError(MSG_ERROR_WHILE_SETTING_PROPERTY.format(
                         prop_id, self.obj_type, self.obj_id, error))

# eof
